from typing import Union


def raise_not_implemented(obj, message: str):
    """
    Raises the ``NotImplementedError`` for convenience
    
    :param obj: 
    :param message: 
    :return: 
    """
    raise NotImplementedError(
        f'{message} is not implemented for objects '
        f'of type {type(obj).__name__}.'
    )


class ChromathicityError(Exception):
    """ Base class for color exceptions"""
    pass


class UndefinedConversionError(ChromathicityError):
    """ Raised when a conversion is undefined """
    def __init__(self, from_space: str, to_space: str):
        self.message = f'Conversion from {from_space} to ' \
                       f'{to_space} is undefined.'


class UndefinedColorSpaceError(ChromathicityError):
    """ Raised when a color space is not recognized """
    def __init__(self, space: Union[str, type]):
        if isinstance(space, str):
            self.message = f'Color space {space} is unrecognized.'
        elif isinstance(space, type):
            self.message = f'Class {space.__name__} is not a recognized ' \
                           f'color space. Use the color_space decorator.'




"""
Keeps all the space names for code consistency. Makes refactoring much easier
"""
REFLECTANCE_SPECTRUM = 'Spectrum'
XYZ = 'CIEXYZ'
XYY = 'xyY'
NORMALIZED_XYZ = 'XYZ_r'
LAB = 'CIELAB'
LCH = 'CIELCH'
LINEAR_RGB = 'lRGB'
RGB = 'RGB'
HSL = 'HSL'
HSI = 'HSI'
HCY = 'HCY'
HSV = 'HSV'






from abc import ABC, abstractmethod

import numpy as np

from chromathicity.illuminant import Illuminant, get_default_illuminant, D
from chromathicity.error import raise_not_implemented
from chromathicity.util import SetGet
from chromathicity.observer import Observer, get_default_observer, Standard


class RgbSpecification(ABC):
    """
    Specifies the RGB Working space
    
    The concept of colorimetrically defined RGB spaces has been around for a long
    time — at least since the days of the development of color television. But the
    popularity in digital imaging applications grew substantially only after Adobe
    introduced the "RGB Working Space" into Photoshop 5.0. Since that time, many
    different working space definitions have been added to the original set of
    color television spaces.
    """

    def __init__(self):
        self.compander = None

    @property
    def name(self):
        return self.__name__

    @name.setter
    def name(self, n):
        raise_not_implemented(self, 'Setting name')

    @property
    def white_point(self):
        return self.illuminant.get_white_point(self.observer)

    @property
    @abstractmethod
    def illuminant(self) -> Illuminant:
        pass

    @illuminant.setter
    def illuminant(self, ill):
        raise_not_implemented(self, 'Setting illuminant')

    @property
    @abstractmethod
    def observer(self) -> Observer:
        pass

    @observer.setter
    def observer(self, obs):
        raise_not_implemented(self, 'Setting setter')

    @property
    @abstractmethod
    def xyy(self) -> np.ndarray:
        pass

    @xyy.setter
    def xyy(self, x):
        raise_not_implemented(self, 'Setting xyy')

    def compand(self, linear_rgb):
        return self.compander.compand(linear_rgb)

    def inverse_compand(self, companded_rgb):
        return self.compander.inverse_compand(companded_rgb)

    @property
    def linear_transformation(self):
        """
        Compute the linear transformation matrix to convert between RGB and XYZ
        
        see http://brucelindbloom.com/index.html?Eqn_XYZ_to_RGB.html for details 
        about the computation. Note that this transformation matrix is the 
        transpose of the one on the website
        :return: 
        """
        # import the convert function here to avoid circular imports
        from chromathicity.convert import convert
        wp = self.white_point
        xyz = convert(self.xyy, 'xyY', 'CIEXYZ', axis=1)
        xyz_normalized = xyz / xyz[:, 1:2]
        s = np.linalg.solve(xyz_normalized.T, wp[:, np.newaxis])
        return s * xyz_normalized


class Custom(RgbSpecification, SetGet):

    def __init__(self, **kwargs):
        super().__init__()
        self._name = ""
        self._illuminant = get_default_illuminant()
        self._observer = get_default_observer()
        self._xyy = np.array([[0.6, 0.3, .200],
                              [0.3, 0.6, .800],
                              [0.2, 0.1, .100]])
        self.set(**kwargs)

    def __repr__(self):
        args = ['name', 'illuminant', 'observer', 'xyy']
        kwargs_repr = ', '.join(f'{key}={getattr(self, key)!r}' for key in args)
        return f'Custom({kwargs_repr!s})'

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, n):
        self._name = n

    @property
    def illuminant(self):
        return self._illuminant

    @illuminant.setter
    def illuminant(self, ill):
        self._illuminant = ill

    @property
    def observer(self):
        return self._observer

    @observer.setter
    def observer(self, obs):
        self._observer = obs

    @property
    def xyy(self):
        return self._xyy

    @xyy.setter
    def xyy(self, x):
        self._xyy = x


class Srgb(RgbSpecification):

    def __init__(self):
        super().__init__()
        self.compander = SrgbCompander()

    def __repr__(self):
        return 'Srgb()'

    @property
    def name(self):
        return 'sRGB'

    @property
    def illuminant(self):
        return D('D_65')

    @property
    def observer(self):
        return Standard(2)

    @property
    def xyy(self):
        return np.array([[0.64, 0.33, .212656],
                         [0.30, 0.60, .715158],
                         [0.15, 0.06,  .072186]])


class Compander(ABC):

    @abstractmethod
    def compand(self, linear_rgb: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def inverse_compand(self, companded_rgb: np.ndarray) -> np.ndarray:
        pass


class SrgbCompander(Compander):

    _EPS = 0.0031308
    _DELTA = 12.92
    _ALPHA = 1.055
    _GAMMA = 2.4
    _BETA = 0.055

    def __repr__(self):
        return 'SrgbCompander()'

    def compand(self, linear_rgb: np.ndarray) -> np.ndarray:
        is_small = linear_rgb <= self._EPS
        is_big = np.logical_not(is_small)
        companded_rgb = np.zeros(linear_rgb.shape)
        companded_rgb[is_small] = self._DELTA * linear_rgb[is_small]
        a = self._ALPHA
        g = self._GAMMA
        b = self._BETA
        companded_rgb[is_big] = a*linear_rgb[is_big] ** (1.0/g) - b
        return companded_rgb

    def inverse_compand(self, companded_rgb: np.ndarray) -> np.ndarray:
        is_small = companded_rgb <= self._DELTA*self._EPS
        is_big = np.logical_not(is_small)
        linear_rgb = np.zeros(companded_rgb.shape)
        linear_rgb[is_small] = companded_rgb[is_small] / self._DELTA
        a = self._ALPHA
        g = self._GAMMA
        b = self._BETA
        linear_rgb[is_big] = ((companded_rgb[is_big] + b) / a) ** g
        return linear_rgb


class GammaCompander(Compander):

    def __init__(self, gamma=1):
        self.gamma = gamma

    def __repr__(self):
        return f'GammaCompander({self.gamma!r})'

    def compand(self, linear_rgb: np.ndarray) -> np.ndarray:
        return linear_rgb ** (1.0 / self.gamma)

    def inverse_compand(self, companded_rgb: np.ndarray) -> np.ndarray:
        return companded_rgb ** self.gamma


def get_default_rgb_specification():
    return Srgb()



from functools import reduce
import operator as op


import numpy as np


def interp1(x, xp, fp):
    """
    Interpolate the 1d piecewise linear interpolant to a function with given 
    values at discrete data points 

    :param x: The x-coordinates of the interpolated values
    :param xp: The x-coordinates of the data points
    :param fp: the y-coordinates of the data points. Multiple functions can be
               interpolated, as long as they lie along the last dimension of 
               the array.
    :return: the interpolated values
    """

    # numpy.interp only accepts 1-D sequences, so all the input data will need
    # to be reshaped
    if isinstance(x, np.ndarray) and x.ndim > 1:
        x = np.reshape(x, -1)
    if isinstance(xp, np.ndarray) and xp.ndim > 1:
        xp = np.reshape(xp, -1)
    y_shape = (*fp.shape[:-1], len(x))
    y = np.zeros(y_shape)
    for i_row in np.ndindex(fp.shape[:-1]):
        y[i_row] = np.interp(x, xp, fp[i_row])
    return y


def is_between(x, a, b, inclusive=(True, True)):
    """Determine the elements that are between a and b"""
    ops = tuple(np.less_equal if inc else np.less for inc in inclusive)
    return np.logical_and(ops[0](a, x), ops[1](x, b))


def product(sequence):
    """
    Computes the product of all elements in the sequence

    :param sequence: an iterable containing things that can be multiplied
    :return: the product
    """
    return reduce(op.mul, sequence, 1)


def to_matrix(array: np.ndarray):
    """reshapes an array so that it is 2-dimensional"""
    if array.ndim == 1:
        return array.reshape(1, array.size)
    elif array.ndim == 2:
        return array.reshape(array.shape)
    else:
        return array.reshape(product(array.shape[:-1]), -1)



from abc import ABC, abstractmethod

import numpy as np


class ChromaticAdaptationAlgorithm(ABC):
    """
    Creates the linear transformations necessary when converting XYZ values 
    between white points. 
    """

    def __repr__(self):
        return f'{type(self).__name__}()'

    def get_linear_transformation(self, white_point_from, white_point_to):
        m_a = self.cone_response_domain
        cone_response_from = white_point_from.reshape((1, 3)).dot(m_a).reshape(-1)
        cone_response_to = white_point_to.reshape((1, 3)).dot(m_a).reshape(-1)
        response_ratio = np.diag(cone_response_to / cone_response_from)
        return np.linalg.solve(m_a.T, m_a.dot(response_ratio).T).T

    @property
    @abstractmethod
    def cone_response_domain(self) -> np.ndarray:
        pass


class Bradford(ChromaticAdaptationAlgorithm):
    """
    This is the current best known chromatic adaptation algorithm
    """

    @property
    def cone_response_domain(self) -> np.ndarray:
        return np.array(
            [
                [0.8951, -0.7502, 0.0389],
                [0.2664, 1.7135, -0.0685],
                [-0.1614, 0.0367, 1.0296]
            ])


class XyzScaling(ChromaticAdaptationAlgorithm):
    """
    This is the "ideal" chromatic adaptation algorithm, but it is the worst.
    """

    @property
    def cone_response_domain(self) -> np.ndarray:
        return np.eye(3)


class VonKries(ChromaticAdaptationAlgorithm):
    """
    Better than XYZ scaling, but worse than Bradford
    """

    @property
    def cone_response_domain(self) -> np.ndarray:
        return np.array(
            [
                [0.40024, -0.2263, 0.0],
                [0.7076, 1.16532, 0.0],
                [-0.08081, 0.0457, 0.91822]
            ])


def get_default_chromatic_adaptation_algorithm():
    return Bradford()



"""
Some of the code in this module was taken and modified from the python-colormath 
package, which is located at 

    https://github.com/gtaylor/python-colormath

Below is the Copyright notice for that code:

Copyright (c) 2014, Greg Taylor
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, 
   this list of conditions and the following disclaimer in the documentation 
   and/or other materials provided with the distribution.
 * Neither the name of the DUO Interactive, LLC nor the names of its 
   contributors may be used to endorse or promote products derived from this 
   software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from abc import ABC, abstractmethod
from functools import wraps
from inspect import signature
from logging import getLogger
from typing import Union, Callable, List, Tuple, Dict, Any, Optional, Type, \
    TYPE_CHECKING

import numpy as np
from networkx import DiGraph, shortest_path, NetworkXNoPath

from chromathicity.error import UndefinedConversionError, \
    UndefinedColorSpaceError
from chromathicity.illuminant import Illuminant
from chromathicity.observer import Observer
from chromathicity.rgbspec import RgbSpecification
from chromathicity.chromadapt import ChromaticAdaptationAlgorithm
import chromathicity.spaces

logger = getLogger(__name__)


class ConversionManager(ABC):

    def __init__(self):
        self.registered_color_spaces = set()

    def add_type_conversion(self, start_type, target_type, conversion_function):
        """
        Register a conversion function between two color spaces.
        :param start_type: Starting color space.
        :param target_type: Target color space.
        :param conversion_function: Conversion function.
        """
        self.registered_color_spaces.add(start_type)
        self.registered_color_spaces.add(target_type)
        logger.debug('Registered conversion from %s to %s', start_type,
                     target_type)

    @abstractmethod
    def get_conversion_path(self, start_space_name, target_space_name):
        """
        Return a list of conversion functions that, if applied iteratively on 
        a color of the ``start_space_name`` color space, result in a color in 
        the ``result_space_name`` color space. Raises an 
        :exc:`UndefinedConversionError` if no valid conversion path can be 
        found. 
        
        :param start_space_name: Starting color space type.
        :param target_space_name: Target color space type.
        :return: List of conversion functions.
        """
        pass


class GraphConversionManager(ConversionManager):
    def __init__(self):
        super().__init__()
        self.conversion_graph = DiGraph()

    def get_conversion_path(self, start_space_name, target_space_name):
        if start_space_name not in self.conversion_graph:
            raise UndefinedColorSpaceError(start_space_name)
        if target_space_name not in self.conversion_graph:
            raise UndefinedColorSpaceError(target_space_name)
        try:
            # Retrieve node sequence that leads from start_space_name to
            # target_space_name.
            path = shortest_path(self.conversion_graph, start_space_name,
                                 target_space_name)
        except NetworkXNoPath:
            raise UndefinedConversionError(
                start_space_name,
                target_space_name,
            )
        # Look up edges between nodes and retrieve the conversion function
        # for each edge.
        cf = 'conversion_function'
        return [self.conversion_graph.get_edge_data(node_a, node_b)[cf]
                for node_a, node_b in zip(path[:-1], path[1:])]

    def add_type_conversion(self, start_space_name, target_space_name,
                            conversion_function):
        super().add_type_conversion(start_space_name,
                                    target_space_name,
                                    conversion_function)
        self.conversion_graph.add_edge(
            start_space_name,
            target_space_name,
            {'conversion_function': conversion_function})


class DummyConversionManager(ConversionManager):
    def add_type_conversion(self, start_space_name, target_space_name,
                            conversion_function):
        pass

    def get_conversion_path(self, start_space_name, target_space_name):
        raise UndefinedConversionError(
            start_space_name,
            target_space_name,
        )


_conversion_manager = GraphConversionManager()

# A type variable for raw conversions before they are passed to the conversion
# decorator.
BareConversion = Callable[[np.ndarray, Any], np.ndarray]

# After applying the color_conversion decorator, bare conversions will have the
# following signature.
Conversion = Callable[
    [
        np.ndarray,
        Any,
        Optional[int],
        Optional[Illuminant],
        Optional[Observer],
        Optional[RgbSpecification],
        Optional[ChromaticAdaptationAlgorithm]
    ],
    np.ndarray]


def color_conversion(from_space_name: str, to_space_name: str) \
        -> Callable[[BareConversion], Conversion]:
    """
    
    Decorator to indicate a function that performs a conversion from one 
    color space to another. This decorator wraps the original function, 
    giving every conversion the same signature:
    
        conversion(data, *args, axis, illuminant, observer, rgbs, caa)
    
    In addition, the function will be registered as a conversion, so it can 
    be used to perform color space transformations between color spaces that 
    do not have direct conversion functions (e.g., Luv to CMYK). The path 
    between two spaces can be found using the :func:`get_conversion_path` 
    function. 
    
    :param from_space_name: Starting color space name or type
    :param to_space_name: Target color space name or type
    """

    def decorator(f: BareConversion) -> Conversion:
        f.start_type = from_space_name
        f.target_type = to_space_name

        f_sig = signature(f)
        kwarg_names = {n for n, p in f_sig.parameters.items()
                       if p.kind == p.KEYWORD_ONLY}
        f._kwarg_names = kwarg_names

        @wraps(f, assigned=('__module__', '__name__', '__qualname__',
                            '__doc__'))
        def convert_wrapper(
                data: np.ndarray,
                *args,
                axis: int=None,
                illuminant: Illuminant=None,
                observer: Observer=None,
                rgbs: RgbSpecification=None,
                caa: ChromaticAdaptationAlgorithm=None) \
                -> np.ndarray:
            local_values = locals()
            kwargs = {n: local_values[n]
                      for n in local_values.keys() & kwarg_names}
            return f(data, *args, **kwargs)

        _conversion_manager.add_type_conversion(from_space_name,
                                                to_space_name,
                                                convert_wrapper)
        return convert_wrapper

    return decorator


def get_conversion_path(from_space: str, to_space: str) -> List[Conversion]:
    """
    Returns a list of functions to apply to perform the conversion. Raises an 
    :class:`UndefinedConversionError` if a conversion path is not found. 
    
    :param from_space: The starting space name
    :param to_space: the destination space name
    """
    return _conversion_manager.get_conversion_path(from_space, to_space)



from abc import ABC, abstractmethod

import numpy as np
import scipy.integrate as integrate
from bidict import bidict

from chromathicity.error import raise_not_implemented
from chromathicity.math import interp1
from chromathicity.util import SetGet
from chromathicity.observer import Observer, Standard, get_default_observer


class Illuminant(ABC):
    """
    Interface for illuminants
    
    A standard illuminant is a theoretical source of visible light with a 
    publish profile (power spectral distribution). Standard illuminants
    provide a basis for comparing images or colors recorded under different
    lighting.
    """

    @property
    def name(self) -> str:
        return self.__name__

    def __str__(self):
        return self.name

    @property
    @abstractmethod
    def wavelengths(self) -> np.ndarray:
        pass

    @property
    @abstractmethod
    def psd(self) -> np.ndarray:
        pass

    def get_psd(self, wavelengths: np.ndarray) -> np.ndarray:
        return interp1(wavelengths, self.wavelengths, self.psd)

    def get_white_point(self, observer: Observer=None) -> np.ndarray:
        """
        Calculate the white point of the illuminant with a specified observer
        
        :param observer: The observer 
        :return: the white point
        """
        if observer is None:
            observer = get_default_observer()
        wls = observer.wavelengths
        p = self.get_psd(wls)
        x_power = observer.xbar * p
        is_valid_x = np.logical_not(np.isnan(x_power))
        x_point = integrate.trapz(x_power[is_valid_x], wls[is_valid_x])
        y_power = observer.ybar * p
        is_valid_y = np.logical_not(np.isnan(y_power))
        y_point = integrate.trapz(y_power[is_valid_y], wls[is_valid_y])
        z_power = observer.zbar * p
        is_valid_z = np.logical_not(np.isnan(z_power))
        z_point = integrate.trapz(z_power[is_valid_z], wls[is_valid_z])
        return np.array([x_point, y_point, z_point])/y_point


class D(Illuminant):
    """
    Natural daylight
    
    Derived by Judd, MacAdam, and Wyszecki, the D series of illuminants are
    constructed to represent natural daylight. They are difficult to produce
    artificially, but are easy to characterize mathematically.
    
    This is, by default, equivalent to D_65.
    """

    _WAVELENGTHS = np.arange(300, 835, 5)
    _S = (
        np.array([
            0.04, 3.02, 6.00, 17.80, 29.60, 42.45, 55.30, 56.30, 57.30, 59.55,
            61.80, 61.65, 61.50, 65.15, 68.80, 66.10, 63.40, 64.60, 65.80,
            80.30, 94.80, 99.80, 104.80, 105.35, 105.90, 101.35, 96.80, 105.35,
            113.90, 119.75, 125.60, 125.55, 125.50, 123.40, 121.30, 121.30,
            121.30, 117.40, 113.50, 113.30, 113.10, 111.95, 110.80, 108.65,
            106.50, 107.65, 108.80, 107.05, 105.30, 104.85, 104.40, 102.20,
            100.00, 98.00, 96.00, 95.55, 95.10, 92.10, 89.10, 89.80, 90.50,
            90.40, 90.30, 89.35, 88.40, 86.20, 84.00, 84.55, 85.10, 83.50,
            81.90, 82.25, 82.60, 83.75, 84.90, 83.10, 81.30, 76.60, 71.90,
            73.10, 74.30, 75.35, 76.40, 69.85, 63.30, 67.50, 71.70, 74.35,
            77.00, 71.10, 65.20, 56.45, 47.70, 58.15, 68.60, 66.80, 65.00,
            65.50, 66.00, 63.50, 61.00, 57.15, 53.30, 56.10, 58.90, 60.40, 61.90
        ]),
        np.array([
            0.02, 2.26, 4.50, 13.45, 22.40, 32.20, 42.00, 41.30, 40.60, 41.10,
            41.60, 39.80, 38.00, 40.20, 42.40, 40.45, 38.50, 36.75, 35.00,
            39.20, 43.40, 44.85, 46.30, 45.10, 43.90, 40.50, 37.10, 36.90,
            36.70, 36.30, 35.90, 34.25, 32.60, 30.25, 27.90, 26.10, 24.30,
            22.20, 20.10, 18.15, 16.20, 14.70, 13.20, 10.90, 8.60, 7.35, 6.10,
            5.15, 4.20, 3.05, 1.90, 0.95, 0.00, -0.80, -1.60, -2.55, -3.50,
            -3.50, -3.50, -4.65, -5.80, -6.50, -7.20, -7.90, -8.60, -9.05,
            -9.50, -10.20, -10.90, -10.80, -10.70, -11.35, -12.00, -13.00,
            -14.00, -13.80, -13.60, -12.80, -12.00, -12.65, -13.30, -13.10,
            -12.90, -11.75, -10.60, -11.10, -11.60, -11.90, -12.20, -11.20,
            -10.20, -9.00, -7.80, -9.50, -11.20, -10.80, -10.40, -10.50,
            -10.60, -10.15, -9.70, -9.00, -8.30, -8.80, -9.30, -9.55, -9.80
        ]),
        np.array([
            0.00, 1.00, 2.00, 3.00, 4.00, 6.25, 8.50, 8.15, 7.80, 7.25, 6.70,
            6.00, 5.30, 5.70, 6.10, 4.55, 3.00, 2.10, 1.20, 0.05, -1.10, -0.80,
            -0.50, -0.60, -0.70, -0.95, -1.20, -1.90, -2.60, -2.75, -2.90,
            -2.85, -2.80, -2.70, -2.60, -2.60, -2.60, -2.20, -1.80, -1.65,
            -1.50, -1.40, -1.30, -1.25, -1.20, -1.10, -1.00, -0.75, -0.50,
            -0.40, -0.30, -0.15, 0.00, 0.10, 0.20, 0.35, 0.50, 1.30, 2.10,
            2.65, 3.20, 3.65, 4.10, 4.40, 4.70, 4.90, 5.10, 5.90, 6.70, 7.00,
            7.30, 7.95, 8.60, 9.20, 9.80, 10.00, 10.20, 9.25, 8.30, 8.95, 9.60,
            9.05, 8.50, 7.75, 7.00, 7.30, 7.60, 7.80, 8.00, 7.35, 6.70, 5.95,
            5.20, 6.30, 7.40, 7.10, 6.80, 6.90, 7.00, 6.70, 6.40, 5.95, 5.50,
            5.80, 6.10, 6.30, 6.50
        ])
    )
    STANDARD_TEMPERATURES = bidict({
        'D_50': 5003,
        'D_55': 5503,
        'D_65': 6504,
        'D_75': 7505
    })

    def __init__(self, temperature=None):
        self._temperature = None
        if temperature is None:
            self._set_temperature('D_65')
        else:
            self._set_temperature(temperature)

    def __repr__(self):
        return 'D({:g})'.format(self.temperature)

    @Illuminant.name.getter
    def name(self):
        if self.temperature in self.STANDARD_TEMPERATURES.inv:
            return self.STANDARD_TEMPERATURES.inv[self.temperature]
        else:
            return "D_{:g}K".format(self.temperature)

    @property
    def temperature(self):
        return self._temperature

    def _set_temperature(self, t):
        if isinstance(t, str):
            if t in self.STANDARD_TEMPERATURES:
                self._temperature = self.STANDARD_TEMPERATURES[t]
            else:
                raise ValueError('Illegal illuminant temperature')
        else:
            if 4000 <= t <= 25000:
                self._temperature = t
            else:
                raise ValueError('Illuminant D must have a temperature between ' 
                                 '4000 and 25000.')

    @Illuminant.wavelengths.getter
    def wavelengths(self):
        if self._WAVELENGTHS.flags.writeable:
            self._WAVELENGTHS.flags.writeable = False
        return self._WAVELENGTHS

    def get_psd(self, wavelengths: np.ndarray):
        s = self._get_daylight_components(wavelengths)
        m = self._daylight_coefficients
        return s[0] + m[0]*s[1] + m[1]*s[2]

    def _get_daylight_components(self, wavelengths):
        return tuple(interp1(wavelengths, self.wavelengths, s) for s in self._S)

    @Illuminant.psd.getter
    def psd(self):
        # noinspection PyTypeChecker
        return self.get_psd(self.wavelengths)

    @property
    def _daylight_coefficients(self):
        x, y = self._chromaticity_coordinates
        md = 0.0241 + 0.2562*x - 0.7341*y
        return ((-1.3515 - 1.7703*x + 5.9114*y)/md,
                (0.0300 - 31.4424*x + 30.0717*y)/md)

    @property
    def _chromaticity_coordinates(self):
        t = self.temperature
        if t <= 7000:
            xd = -4.607e9/t**3 + 2.9678e6/t**2 + 0.09911e3/t + 0.244063
        else:
            xd = -2.0064e9/t**3 + 1.9018e6/t**2 + 0.24748e3/t + 0.237040

        return xd, -3*xd**2 + 2.87*xd - 0.275


class A(Illuminant):
    """
    Standard illuminant based on a Planckian Locus
    
    The CIE defines illuminant A in these terms:
    
      CIE standard illuminant A is intended to represent typical, domestic, 
      tungsten-filament lighting. Its relative spectral power distribution is 
      that of a Planckian radiator at a temperature of approximately 2856 K. 
      CIE standard illuminant A should be used in all applications of 
      colorimetry involving the use of incandescent lighting, unless there 
      are specific reasons for using a different illuminant. 
    
          —CIE, CIE Standard Illuminants for Colorimetry
    
    This class is more general than the CIE standard illuminant A in that you 
    can change the color temperature 
    """

    _DEFAULT_TEMPERATURE = 2848.0
    _WAVELENGTHS = np.arange(300.0, 831.0)
    _C_1 = 560.0
    _C_2 = 1.435e7

    def __init__(self, temperature: float=None):
        self.temperature = self._DEFAULT_TEMPERATURE
        if temperature is not None:
            self.temperature = temperature

    def __repr__(self):
        return 'A({})'.format(self.temperature)

    @Illuminant.name.getter
    def name(self):
        if self.temperature == self._DEFAULT_TEMPERATURE:
            return 'A'
        else:
            return 'A_{}K'.format(self.temperature)

    @Illuminant.wavelengths.getter
    def wavelengths(self):
        return self._WAVELENGTHS

    @Illuminant.psd.getter
    def psd(self):
        # noinspection PyTypeChecker
        return self.get_psd(self.wavelengths)

    def get_psd(self, wavelengths: np.ndarray=None) -> np.ndarray:
        if wavelengths is None:
            wavelengths = self.wavelengths
        t = self.temperature
        wl = wavelengths
        c1 = self._C_1
        c2 = self._C_2
        return (
            100 * (c1 / wl)**5.0 * (np.exp(c2 / (t*c1)) - 1)
            / (np.exp(c2/(t*wl)) - 1)
        )


class E(Illuminant):
    """
    Ideal Illuminant
    
    From `Wikipedia <https://en.wikipedia.org/wiki/
    Standard_illuminant#Illuminant_E>`_:
    
        *Illuminant E is an equal-energy radiator; it has a constant SPD 
        inside the visible spectrum. It is useful as a theoretical reference; 
        an illuminant that gives equal weight to all wavelengths, presenting 
        an even color. It also has equal CIE XYZ tristimulus values, thus its 
        chromaticity coordinates are (x,y)=(1/3,1/3). This is by design; the 
        XYZ color matching functions are normalized such that their integrals 
        over the visible spectrum are the same.*

        *Illuminant E is not a black body, so it does not have a color 
        temperature, but it can be approximated by a D series illuminant with 
        a CCT of 5455 K. (Of the canonical illuminants, D55 is the closest.) 
        Manufacturers sometimes compare light sources against Illuminant E to 
        calculate the excitation purity.* 
    """
    _WAVELENGTHS = np.arange(360, 10, 831)

    def __repr__(self):
        return 'E()'

    @Illuminant.wavelengths.getter
    def wavelengths(self):
        if self._WAVELENGTHS.flags.writeable:
            self._WAVELENGTHS.flags.writeable = False
        return self._WAVELENGTHS

    @Illuminant.psd.getter
    def psd(self):
        # noinspection PyTypeChecker
        return self.get_psd(self.wavelengths)

    def get_psd(self, wavelengths: np.ndarray):
        return np.ones((len(wavelengths),))


# noinspection PyMethodOverriding
class CustomIlluminant(Illuminant):
    """
    A class that allows dynamic creation of new illuminants
    """

    def __init__(self, *, name='', wavelengths=(), psd=()):
        self._name = name
        self._wavelengths = np.array(wavelengths)
        self._psd = np.array(psd)

    @Illuminant.name.getter
    def name(self):
        return self._name

    @Illuminant.wavelengths.getter
    def wavelengths(self):
        return self._wavelengths

    @Illuminant.psd.getter
    def psd(self):
        return self._psd


def get_default_illuminant() -> Illuminant:
    return D()



from typing import Tuple

import numpy as np
import scipy.integrate as integrate

from chromathicity.chromadapt import (
    ChromaticAdaptationAlgorithm,
    get_default_chromatic_adaptation_algorithm)
from chromathicity.illuminant import Illuminant, get_default_illuminant
from chromathicity.manage import color_conversion, get_conversion_path
from chromathicity.observer import Observer, get_default_observer
from chromathicity.rgbspec import (RgbSpecification,
                                   get_default_rgb_specification)
import chromathicity.space_names as names
# Constants used for L*a*b* conversions
from chromathicity.util import construct_component_inds, get_matching_axis

LAB_EPS = 0.008856
LAB_KAPPA = 903.3


def convert(data: np.ndarray, from_space: str, to_space: str, *args,
            **kwargs) -> np.ndarray:
    """
    Convert data between spaces


    """
    conversion_path = get_conversion_path(from_space, to_space)
    for do_conversion_step in conversion_path:
        data = do_conversion_step(data, *args, **kwargs)
    return data


@color_conversion(names.LAB, names.NORMALIZED_XYZ)
def lab2xyzr(lab: np.ndarray, *, axis: int=None) -> np.ndarray:
    """
    Convert LAB to normalized XYZ
    
    :param lab: 
    :param axis: 
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(lab.shape, 3)

    inds = construct_component_inds(axis, len(lab.shape), 3)
    fxyz = np.zeros(lab.shape)
    fxyz[inds[1]] = (lab[inds[0]] + 16) / 116
    fxyz[inds[0]] = lab[inds[1]]/500 + fxyz[inds[1]]
    fxyz[inds[2]] = fxyz[inds[1]] - lab[inds[2]]/200

    is_small = fxyz <= (LAB_EPS ** (1.0/3.0))
    is_big = np.logical_not(is_small)
    xyzr = np.zeros(lab.shape)
    xyzr[is_big] = fxyz[is_big] ** 3.0
    xyzr[is_small] = (116*fxyz[is_small] - 16) / LAB_KAPPA
    return xyzr


@color_conversion(names.LAB, names.LCH)
def lab2lch(lab: np.ndarray, *, axis: int=None) -> np.ndarray:
    """
    Convert L*a*b* to LCh
    
    :param lab: 
    :param axis: 
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(lab.shape, 3)

    inds = construct_component_inds(axis, lab.ndim, 3)

    lch = np.zeros(lab.shape)

    lch[inds[0]] = lab[inds[0]]
    lch[inds[1]] = np.sqrt(lab[inds[1]]**2 + lab[inds[2]]**2)
    lch[inds[2]] = np.mod((180/np.pi)*np.arctan2(lab[inds[2]], lab[inds[1]]),
                          360.)

    return lch


@color_conversion(names.LCH, names.LAB)
def lch2lab(lch: np.ndarray, *, axis: int=None) -> np.ndarray:
    """
    Converts LCh to L*a*b*
    
    :param lch: 
    :param axis: 
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(lch.shape, 3)

    inds = construct_component_inds(axis, lch.ndim, 3)

    lab = np.zeros(lch.shape)
    lab[inds[0]] = lch[inds[0]]
    h_rad = (np.pi/180)*lch[inds[2]]
    lab[inds[1]] = lch[inds[1]]*np.cos(h_rad)
    lab[inds[2]] = lch[inds[1]]*np.sin(h_rad)

    return lab


@color_conversion(names.LINEAR_RGB, names.RGB)
def lrgb2rgb(lrgb: np.ndarray, *, rgbs: RgbSpecification=None) -> np.ndarray:
    """
    Convert linear RGB to companded RGB
    
    :param lrgb: 
    :param rgbs: 
    :return: 
    """
    if rgbs is None:
        rgbs = get_default_rgb_specification()
    return rgbs.compand(lrgb)


@color_conversion(names.LINEAR_RGB, names.XYZ)
def lrgb2xyz(lrgb: np.ndarray,
             *,
             axis: int=None,
             illuminant: Illuminant=None,
             observer: Observer=None,
             rgbs: RgbSpecification=None,
             caa: ChromaticAdaptationAlgorithm=None) -> np.ndarray:
    """
    Convert from linear RGB to XYZ
    
    :param lrgb: 
    :param axis: 
    :param illuminant: 
    :param observer: 
    :param rgbs: 
    :param caa: 
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(lrgb.shape, 3)
    if illuminant is None:
        illuminant = get_default_illuminant()
    if observer is None:
        observer = get_default_observer()
    if rgbs is None:
        rgbs = get_default_rgb_specification()
    if caa is None:
        caa = get_default_chromatic_adaptation_algorithm()

    # Get the XYZ values into the correct shape for matrix multiplication
    n_dims = lrgb.ndim
    new_dims = list(range(n_dims))
    if axis != n_dims - 1:
        new_dims[-1] = axis
        new_dims[axis] = n_dims - 1
    lrgb = lrgb.transpose(new_dims)

    input_shape = lrgb.shape
    lrgb_is_not_matrix = n_dims != 2
    if lrgb_is_not_matrix:
        lrgb = lrgb.reshape((-1, 3))

    # Do the transformation
    m = rgbs.linear_transformation
    xyz = lrgb.dot(m)

    # Transform back to the original shape
    if lrgb_is_not_matrix:
        xyz = xyz.reshape(input_shape)

    if axis != n_dims - 1:
        xyz = xyz.transpose(new_dims)

    source_white_point = rgbs.white_point
    destination_white_point = illuminant.get_white_point(observer)
    if not np.allclose(source_white_point, destination_white_point,
                       rtol=1e-5, atol=1e-14):
        return xyz2xyz(xyz, source_white_point, destination_white_point, axis,
                       caa)
    else:
        return xyz


@color_conversion(names.RGB, names.LINEAR_RGB)
def rgb2lrgb(rgb: np.ndarray,
             *,
             rgbs: RgbSpecification=None) -> np.ndarray:
    """
    Convert companded RGB to linear RGB
    
    :param rgb: 
    :param rgbs: 
    :return: 
    """
    if rgbs is None:
        rgbs = get_default_rgb_specification()
    return rgbs.inverse_compand(rgb)


@color_conversion(names.RGB, names.HSL)
def rgb2hsl(rgb: np.ndarray,
            *,
            axis: int=None) -> np.ndarray:
    """
    Convert RGB to Hue Saturation Lightness
    :param rgb: 
    :param axis: 
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(rgb.shape, 3)
    big_m, little_m, chroma = _compute_chroma(rgb, axis)

    inds = construct_component_inds(axis, rgb.ndim, 3)

    hsl = np.zeros(rgb.shape)
    hsl[inds[0]] = _compute_rgb_hue(rgb, big_m, little_m, chroma, axis)
    l = 0.5 * (big_m + little_m)
    hsl[inds[2]] = l

    l_lt_one = l < 1.
    if np.any(l_lt_one):
        hsl[inds[1]][l_lt_one] = (
            (big_m[l_lt_one] - little_m[l_lt_one])
            / (1. - np.abs(2*l[l_lt_one] - 1)))

    return hsl


@color_conversion(names.HSL, names.RGB)
def hsl2rgb(hsl: np.ndarray,
            *,
            axis: int=None) -> np.ndarray:
    """
    Convert from Hue Saturation Lightness (HSL) to RGB
    
    :type hsl: np.ndarray
    :param hsl: 
    :param axis: 
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(hsl.shape, 3)

    inds = construct_component_inds(axis, hsl.ndim, 3)

    chroma = ((1. - np.abs(2. * hsl[inds[2]] - 1.))
              * hsl[inds[1]])  # type: np.ndarray

    h_prime = hsl[inds[0]] / 60.
    x = chroma * (1 - np.abs(np.mod(h_prime, 2) - 1))
    rgb1 = _compute_rgb1(hsl.shape, inds, h_prime, x, chroma)

    little_m = hsl[inds[2]] - chroma/2.
    if little_m.ndim > rgb1.ndim:
        # This only happens if hsl is 1-D
        return rgb1 + little_m[0]
    else:
        return rgb1 + little_m


@color_conversion(names.RGB, names.HSI)
def rgb2hsi(rgb: np.ndarray,
            *,
            axis: int=None) -> np.ndarray:
    """
    Convert RGB to Hue Saturation Intensity
    
    :param rgb: 
    :param axis: 
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(rgb.shape, 3)
    big_m, little_m, chroma = _compute_chroma(rgb, axis)

    inds = construct_component_inds(axis, rgb.ndim, 3)

    hsi = np.zeros(rgb.shape)
    hsi[inds[0]] = _compute_rgb_hue(rgb, big_m, little_m, chroma, axis)
    hsi[inds[2]] = np.mean(rgb, axis=axis, keepdims=True)

    i_nz = hsi[inds[2]] != 0  # type: np.ndarray
    if little_m.ndim < i_nz.ndim:
        # This only happens in the 1D case
        little_m = little_m[slice(None), np.newaxis]
    if np.any(i_nz):
        hsi[inds[1]][i_nz] = 1 - little_m[i_nz] / hsi[inds[2]][i_nz]

    return hsi


@color_conversion(names.HSI, names.RGB)
def hsi2rgb(hsi: np.ndarray,
            *,
            axis: int=None) -> np.ndarray:
    """
    Convert Hue Saturation Intensity (HSI) to RGB
    
    :param hsi: 
    :param axis: 
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(hsi.shape, 3)

    inds = construct_component_inds(axis, hsi.ndim, 3)

    h_prime = hsi[inds[0]] / 60.
    z = 1 - np.abs(np.mod(h_prime, 2.) - 1)
    chroma = 3*hsi[inds[2]]*hsi[inds[1]]/(1 + z)  # type: np.ndarray
    x = chroma * z
    rgb1 = _compute_rgb1(hsi.shape, inds, h_prime, x, chroma)
    little_m = hsi[inds[2]] * (1 - hsi[inds[1]])  # type: np.ndarray
    if little_m.ndim > rgb1.ndim:
        # This only happens in the 1D case
        return rgb1 + little_m[0]
    else:
        return rgb1 + little_m


@color_conversion(names.RGB, names.HSV)
def rgb2hsv(rgb: np.ndarray, *, axis: int=None) -> np.ndarray:
    """
    Convert from RGB to Hue Saturation Value (HSV)
    
    :param rgb: 
    :param axis: 
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(rgb.shape, 3)

    big_m, little_m, chroma = _compute_chroma(rgb, axis)

    inds = construct_component_inds(axis, rgb.ndim, 3)

    hsv = np.zeros(rgb.shape)
    hsv[inds[0]] = _compute_rgb_hue(rgb, big_m, little_m, chroma, axis)
    hsv[inds[2]] = big_m

    big_m_nz = big_m != 0
    hsv[inds[1]][big_m_nz] = chroma[big_m_nz] / big_m[big_m_nz]

    return hsv


@color_conversion(names.HSV, names.RGB)
def hsv2rgb(hsv: np.ndarray, *, axis: int=None) -> np.ndarray:
    """
    Convert from HSV to RGB
    :param hsv: 
    :param axis: 
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(hsv.shape, 3)

    inds = construct_component_inds(axis, hsv.ndim, 3)

    chroma = hsv[inds[1]] * hsv[inds[2]]
    h_prime = hsv[inds[0]] / 60.
    x = chroma * (1. - np.abs(np.mod(h_prime, 2.) - 1))  # type: np.ndarray
    rgb1 = _compute_rgb1(hsv.shape, inds, h_prime, x, chroma)
    little_m = hsv[inds[2]] - chroma

    if little_m.ndim > rgb1.ndim:
        # This only happens in the 1D case
        return rgb1 + little_m[0]
    else:
        return rgb1 + little_m


@color_conversion(names.RGB, names.HCY)
def rgb2hcy(rgb: np.ndarray, *, axis: int=None) -> np.ndarray:
    """
    Convert from RGB to Hue, Chroma, Luma (Y'_601)
    
    :param rgb: 
    :param axis: 
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(rgb.shape, 3)

    big_m, little_m, chroma = _compute_chroma(rgb, axis)

    inds = construct_component_inds(axis, rgb.ndim, 3)

    hcy = np.zeros(rgb.shape)
    hcy[inds[0]] = _compute_rgb_hue(rgb, big_m, little_m, chroma, axis)
    hcy[inds[1]] = chroma
    hcy[inds[2]] = 0.299*rgb[inds[0]] + 0.587*rgb[inds[1]] + 0.114*rgb[inds[2]]

    return hcy


@color_conversion(names.HCY, names.RGB)
def hcy2rgb(hcy: np.ndarray, *, axis: int=None) -> np.ndarray:
    """
    
    :param hcy: 
    :param axis: 
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(hcy.shape, 3)

    inds = construct_component_inds(axis, hcy.ndim, 3)

    h_prime = hcy[inds[0]] / 60.
    x: np.ndarray = hcy[inds[1]] * (1 - np.abs(np.mod(h_prime, 2) - 1))
    rgb1 = _compute_rgb1(hcy.shape, inds, h_prime, x, hcy[inds[1]])
    little_m: np.ndarray = hcy[inds[2]] - (0.299*rgb1[inds[0]]
                                           + 0.587*rgb1[inds[1]]
                                           + 0.114*rgb1[inds[2]])

    if little_m.ndim > rgb1.ndim:
        # This only happens in the 1D case
        return rgb1 + little_m[0]
    else:
        return rgb1 + little_m


def _compute_chroma(rgb: np.ndarray,
                    axis: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Compute the chroma for an RGB triple"""
    big_m = np.max(rgb, axis=axis, keepdims=True)
    little_m = np.min(rgb, axis=axis, keepdims=True)
    chroma = big_m - little_m
    return big_m, little_m, chroma


def _compute_rgb_hue(rgb: np.ndarray, big_m, little_m, chroma,
                     axis: int) -> np.ndarray:
    """ Compute the RGB Hue. This is the same for HSV and HSL """
    inds = construct_component_inds(axis, rgb.ndim, 3)
    r = rgb[inds[0]]
    g = rgb[inds[1]]
    b = rgb[inds[2]]
    if chroma.ndim < r.ndim:
        # this will only happen when chroma is 1-D
        chroma = chroma[(slice(None), np.newaxis)]
    ch_is_nz = chroma != 0.
    m_is_r = np.logical_and(big_m == r, ch_is_nz)
    m_is_g = np.logical_and(big_m == g, ch_is_nz)
    m_is_b = np.logical_and(big_m == b, ch_is_nz)

    h_prime = np.zeros(r.shape)
    if np.any(m_is_r):
        h_prime[m_is_r] = np.mod((g[m_is_r] - b[m_is_r]) / chroma[m_is_r] + 6.,
                                 6.)
    if np.any(m_is_g):
        h_prime[m_is_g] = (b[m_is_g] - r[m_is_g]) / chroma[m_is_g] + 2.
    if np.any(m_is_b):
        h_prime[m_is_b] = (r[m_is_b] - g[m_is_b]) / chroma[m_is_b] + 4.
    return 60. * h_prime


def _compute_rgb1(shape, inds, h_prime, x, chroma) -> np.ndarray:
    rgb1 = np.zeros(shape)
    i01 = np.logical_and(0. <= h_prime, h_prime <= 1.)
    i12 = np.logical_and(1. <= h_prime, h_prime <= 2.)
    i23 = np.logical_and(2. <= h_prime, h_prime <= 3.)
    i34 = np.logical_and(3. <= h_prime, h_prime <= 4.)
    i45 = np.logical_and(4. <= h_prime, h_prime <= 5.)
    i56 = np.logical_and(5. <= h_prime, h_prime <= 6.)
    or_ = np.logical_or
    rgb1[inds[0]][or_(i01, i56)] = chroma[or_(i01, i56)]
    rgb1[inds[0]][or_(i12, i45)] = x[or_(i12, i45)]
    rgb1[inds[1]][or_(i01, i34)] = x[or_(i01, i34)]
    rgb1[inds[1]][or_(i12, i23)] = chroma[or_(i12, i23)]
    rgb1[inds[2]][or_(i23, i56)] = x[or_(i23, i56)]
    rgb1[inds[2]][or_(i34, i45)] = chroma[or_(i34, i45)]

    return rgb1


@color_conversion(names.REFLECTANCE_SPECTRUM, names.XYZ)
def spectrum2xyz(spectrum: np.ndarray,
                 wavelengths: np.ndarray,
                 *,
                 axis: int=None,
                 illuminant: Illuminant=get_default_illuminant(),
                 observer: Observer=get_default_observer()) -> np.ndarray:
    """
    Convert reflectance spectrum to XYZ
    
    :param spectrum: the reflectance spectrum
    :param wavelengths: the wavelengths corresponding to the spectra
    :param axis: The axis along which the spectra lie. If this is None, 
        then the axis is the last axis with a size that matches wavelengths. 
    :param illuminant: the illuminant
    :param observer: the observer
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(spectrum.shape, wavelengths.size)

    # Need to ensure that we reshape all the 1-D parts of the integral, so that
    # they can be broadcast properly
    new_shape = [1] * len(spectrum.shape)
    new_shape[axis] = -1
    wavelengths = wavelengths.reshape(new_shape)
    xbar = observer.get_xbar(wavelengths).reshape(new_shape)
    ybar = observer.get_ybar(wavelengths).reshape(new_shape)
    zbar = observer.get_zbar(wavelengths).reshape(new_shape)

    illuminant_psd = illuminant.get_psd(wavelengths).reshape(new_shape)

    norm_factor = (1
                   / integrate.trapz(illuminant_psd * ybar,
                                     x=wavelengths, axis=axis))

    phi = spectrum * illuminant_psd

    # This is the shape that each XYZ component must be
    component_shape = list(spectrum.shape)
    component_shape[axis] = 1

    def integrate_phi(bar: np.ndarray):
        area = norm_factor * integrate.trapz(bar * phi, wavelengths, axis=axis)
        # noinspection PyTypeChecker
        return np.reshape(area, component_shape)

    x = integrate_phi(xbar)
    y = integrate_phi(ybar)
    z = integrate_phi(zbar)
    return np.concatenate((x, y, z), axis=axis)


@color_conversion(names.XYY, names.XYZ)
def xyy2xyz(xyy, *, axis: int=None) -> np.ndarray:
    """
    converts from xyY to XYZ
    
    :param xyy: 
    :param axis: 
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(xyy.shape, 3)

    inds = construct_component_inds(axis, len(xyy.shape), 3)

    # Determine where y iz 0 so we don't divide by it
    nzy = np.nonzero(xyy[inds[1]])
    xyz = np.zeros(xyy.shape)

    xyz[inds[0]][nzy] = xyy[inds[0]][nzy]*xyy[inds[2]][nzy] / xyy[inds[1]][nzy]
    xyz[inds[1]][nzy] = xyy[inds[2]][nzy]
    xyz[inds[2]][nzy] = (
        (1 - xyy[inds[0]][nzy] - xyy[inds[1]][nzy]) * xyy[inds[2]][nzy]
        / xyy[inds[1]][nzy]
    )
    return xyz


@color_conversion(names.XYZ, names.LINEAR_RGB)
def xyz2lrgb(xyz: np.ndarray, *,
             axis: int=None,
             illuminant: Illuminant=None,
             observer: Observer=None,
             rgbs: RgbSpecification=None,
             caa: ChromaticAdaptationAlgorithm=None) -> np.ndarray:
    """
    Convert XYZ to linear RGB
    
    :param xyz: 
    :param axis: 
    :param illuminant: 
    :param observer: 
    :param rgbs: 
    :param caa: 
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(xyz.shape, 3)
    if illuminant is None:
        illuminant = get_default_illuminant()
    if observer is None:
        observer = get_default_observer()
    if rgbs is None:
        rgbs = get_default_rgb_specification()
    if caa is None:
        caa = get_default_chromatic_adaptation_algorithm()

    # If the white points are not equal, we will need to convert to the
    # RGB white point
    source_white_point = illuminant.get_white_point(observer)
    destination_white_point = rgbs.white_point
    if not np.allclose(source_white_point, destination_white_point,
                       rtol=1e-5, atol=1e-14):
        xyz = xyz2xyz(xyz, source_white_point, destination_white_point, axis,
                      caa)

    # Get the XYZ values into the correct shape for matrix multiplication
    n_dims = xyz.ndim
    new_dims = list(range(n_dims))
    if axis != n_dims - 1:
        new_dims[-1] = axis
        new_dims[axis] = n_dims - 1
    xyz = xyz.transpose(new_dims)

    input_shape = xyz.shape
    xyz_is_not_matrix = xyz.ndim != 2
    if xyz_is_not_matrix:
        xyz = xyz.reshape((-1, 3))

    # Convert to linear RGB
    m = rgbs.linear_transformation
    lrgb = np.linalg.solve(m.T, xyz.T).T

    # Transform the destination data back to the original shape
    if xyz_is_not_matrix:
        lrgb = lrgb.reshape(input_shape)

    return lrgb.transpose(new_dims)


@color_conversion(names.XYZ, names.XYY)
def xyz2xyy(xyz: np.ndarray, *,
            axis: int=None,
            illuminant: Illuminant=None,
            observer: Observer=None) -> np.ndarray:
    """
    Convert XYZ to xyY
    
    :param xyz: 
    :param axis: 
    :param illuminant: 
    :param observer: 
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(xyz.shape, 3)
    if illuminant is None:
        illuminant = get_default_illuminant()
    if observer is None:
        observer = get_default_observer()
    inds = construct_component_inds(axis, xyz.ndim, 3)

    denominator = np.sum(xyz, axis, keepdims=True)
    nzd = denominator != 0

    xyy = np.zeros(xyz.shape)
    if np.any(nzd):
        xyy[inds[0]][nzd] = xyz[inds[0]][nzd] / denominator[nzd]
        xyy[inds[1]][nzd] = xyz[inds[1]][nzd] / denominator[nzd]
    xyy[inds[2]] = xyz[inds[1]]
    if not np.all(nzd):
        # For any point that is pure black (X=Y=Z=0), give it the
        # chromaticity of the white point of the specified illuminant and
        # observer.
        white_point = illuminant.get_white_point(observer)
        # to prevent infinite recursion, ensure that the white point is
        # non-black
        if white_point[1] > 0:
            zd = np.logical_not(nzd)
            white_point_xyy = xyz2xyy(white_point,
                                      illuminant=illuminant,
                                      observer=observer)
            xyy[inds[0]][zd] = white_point_xyy[0]
            xyy[inds[1]][zd] = white_point_xyy[1]
    return xyy


def xyz2xyz(source_xyz: np.ndarray,
            source_white_point: np.ndarray,
            destination_white_point: np.ndarray,
            axis: int=None,
            caa: ChromaticAdaptationAlgorithm=None) -> np.ndarray:
    """
    Convert XYZ values between two white points
    
    :param source_xyz: 
    :param source_white_point: 
    :param destination_white_point: 
    :param axis: 
    :param caa: 
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(source_xyz.shape, 3)
    if caa is None:
        caa = get_default_chromatic_adaptation_algorithm()

    # Get the source color into the correct shape for matrix multiplication
    n_dims = source_xyz.ndim

    new_dims = list(range(n_dims))
    if axis != n_dims - 1:
        new_dims[-1] = axis
        new_dims[axis] = n_dims - 1
    source_xyz = source_xyz.transpose(new_dims)

    input_shape = source_xyz.shape
    xyz_is_not_matrix = n_dims != 2
    if xyz_is_not_matrix:
        source_xyz = source_xyz.reshape((-1, 3))

    # Convert between the white points
    m = caa.get_linear_transformation(source_white_point,
                                      destination_white_point)
    destination_xyz = source_xyz.dot(m)
    # Now convert the shape back to the original
    if xyz_is_not_matrix:
        destination_xyz = destination_xyz.reshape(input_shape)

    return destination_xyz.transpose(new_dims)


@color_conversion(names.XYZ, names.NORMALIZED_XYZ)
def xyz2xyzr(xyz: np.ndarray, *,
             axis: int=None,
             illuminant: Illuminant=get_default_illuminant(),
             observer: Observer=get_default_observer()) -> np.ndarray:
    """
    Convert XYZ to normalized XYZ reflectance
    :param xyz: the raw xyz values
    :param axis: the axis that the XYZ values lie along 
    :param illuminant: the illuminant
    :param observer: the observer
    :return: the xyz normalized Reflectance
    """
    if axis is None:
        axis = get_matching_axis(xyz.shape, 3)

    new_shape = [1] * len(xyz.shape)
    new_shape[axis] = -1
    white_point = illuminant.get_white_point(observer).reshape(new_shape)
    return xyz / white_point


@color_conversion(names.NORMALIZED_XYZ, names.LAB)
def xyzr2lab(xyzr: np.ndarray, *,
             axis: int=None) -> np.ndarray:
    """
    Convert from normalized XYZ to LAB
    
    :param xyzr: normalized XYZ
    :param axis: the axis along which the values lie
    :return: LAB
    """
    if axis is None:
        axis = get_matching_axis(xyzr.shape, 3)

    fxyz = np.zeros(xyzr.shape)
    is_big = xyzr > LAB_EPS
    is_small = xyzr <= LAB_EPS
    fxyz[is_big] = np.power(xyzr[is_big], 1.0/3.0)
    fxyz[is_small] = (LAB_KAPPA * xyzr[is_small] + 16) / 116

    lab = np.zeros(xyzr.shape)
    # Construct the indices for the 3 components
    inds = construct_component_inds(axis, len(xyzr.shape), 3)
    lab[inds[0]] = 116*fxyz[inds[1]] - 16
    lab[inds[1]] = 500*(fxyz[inds[0]] - fxyz[inds[1]])
    lab[inds[2]] = 200*(fxyz[inds[1]] - fxyz[inds[2]])
    return lab


@color_conversion(names.NORMALIZED_XYZ, names.XYZ)
def xyzr2xyz(xyzr: np.ndarray, *,
             axis: int=None,
             illuminant: Illuminant=None,
             observer: Observer=None) -> np.ndarray:
    """
    Convert normalized XYZ to LAB

    :param xyzr:
    :param axis: 
    :param illuminant: 
    :param observer: 
    :return: 
    """
    if axis is None:
        axis = get_matching_axis(xyzr.shape, 3)

    if illuminant is None:
        illuminant = get_default_illuminant()

    if observer is None:
        observer = get_default_observer()

    new_shape = tuple(-1 if dim == axis else 1
                      for dim in range(len(xyzr.shape)))
    white_point = illuminant.get_white_point(observer).reshape(new_shape)
    return xyzr * white_point



from typing import Any, Iterable, Tuple, TypeVar, Generic, Callable, Union

import numpy as np


class SetGet:
    """
    Provides a useful set and get interface
    """

    def set(self, **kwargs):
        """
        Set attributes::
        
            obj.set(attr1='value', attr2=35, attr3=True)
        
        :param kwargs: ``name=value`` pairs of attributes to set
        :return: self
        """
        for key in kwargs:
            setattr(self, key, kwargs[key])
        return self

    def get(self, *args) -> Iterable[Any]:
        """
        Get a number of attributes::
        
            obj.get('attr1', 'attr2', 'attr3')
        
        :param args: a number of attribute names to return
        :return: An iterable containing the attributes
        """
        return (getattr(self, key) for key in args)


def construct_component_inds(axis: int,
                             n_dims: int,
                             n_components: int,
                             min_ndims: int=2) -> Tuple[Tuple]:
    """
    Construct a tuple of tuples, where each element extracts the correct 
    component values.
    
    :param axis: 
    :param n_dims: 
    :param n_components: 
    :param min_ndims:
    :return: 
    """
    # noinspection PyTypeChecker
    return tuple(
        tuple(slice(i, i+1)
              if dim == axis
              else (slice(None) if dim < n_dims else np.newaxis)
              for dim in range(max(n_dims, min_ndims)))
        for i in range(n_components))


def get_matching_axis(shape: Tuple, length: int) -> int:
    """
    Infers the correct axis to use
    
    :param shape: the shape of the input
    :param length: the desired length of the axis
    :return: the correct axis. If multiple axes match, then it returns the last 
             one.
    """
    # noinspection PyUnresolvedReferences
    axis_candidates = np.nonzero(np.array(shape) == length)[0]
    if len(axis_candidates) == 0:
        raise ValueError('Unable to infer axis tue to shape mismatch: ' 
                         '{} =/= {}.'.format(shape, length))
    return axis_candidates[-1]

A = TypeVar('A')
T = TypeVar('T')
GetMethod = Callable[..., A]
SetMethod = Callable[[T, A], None]
DelMethod = Callable[[T], None]


# noinspection PyPep8Naming
class lazy_property(Generic[A]):
    """
    A property-like descriptor that does not bind to a function, but to the name 
    of the function. That way subclasses can easily override the getter/setter/
    delete
    """
    def __init__(self,
                 getter_method: GetMethod=None,
                 setter_method: SetMethod=None,
                 deleter_method: DelMethod=None,
                 doc=None):
        self.getter_method = getter_method
        self.setter_method = setter_method
        self.deleter_method = deleter_method
        if doc is None:
            if getter_method.__doc__ is not None:
                doc = getter_method.__doc__
            elif setter_method is not None:
                doc = setter_method.__doc__
        self.__doc__ = doc

    def __get__(self, obj, cls=None) -> Union['lazy_property', A]:
        if obj is None:
            return self
        if self.getter_method is None:
            raise AttributeError('unreadable attribute')
        try:
            fget = getattr(obj, self.getter_method.__name__)
        except AttributeError:
            raise TypeError(f'{type(obj).__name__} object does not have '
                            f'a {self.getter_method.__name__} method')
        return fget()

    def __set__(self, obj, value: A):
        if self.setter_method is None:
            raise AttributeError("can't set attribute")
        try:
            fset = getattr(obj, self.setter_method.__name__)
        except AttributeError:
            raise TypeError(f'{type(obj).__name__} object does not have '
                            f'a {self.setter_method.__name__} method.')
        fset(value)

    def __delete__(self, obj):
        if self.deleter_method is None:
            raise AttributeError("can't delete attribute")
        try:
            fdel = getattr(obj, self.deleter_method.__name__)
        except AttributeError:
            raise TypeError(f'{type(obj).__name__} object does not have '
                            f'a {self.deleter_method.__name__} method')
        fdel()



"""
Color space objects
===================
   
"""


from abc import ABC, abstractmethod
from typing import Union, Iterable, Tuple, Any, Type, Dict, Callable

import numpy as np

import chromathicity.space_names as names
from chromathicity.chromadapt import (
    get_default_chromatic_adaptation_algorithm,
    ChromaticAdaptationAlgorithm)
from chromathicity.error import UndefinedColorSpaceError
from chromathicity.illuminant import get_default_illuminant, Illuminant
from chromathicity.observer import get_default_observer, Observer
from chromathicity.rgbspec import (get_default_rgb_specification,
                                   RgbSpecification)
from chromathicity.util import SetGet, construct_component_inds, \
    get_matching_axis, lazy_property


class ColorSpaceData(ABC):
    """
    Defines the main interface for all color space data classes. 
    :class:`ColorSpaceDataImpl` provides a full implementation of this 
    interface, which makes it much easier to define custom color spaces.
    
    Since this class is designed to be extended, every property calls a 
    respective getter and setter method, so that overriding is much easier
    in subclasses.
    """

    # Controls how scaling works. If :attr:`~ColorSpaceDataImpl.is_scaled` is
    # ``True``, then the data will be scaled by this value
    max_value: np.ndarray = [np.inf]

    # The minimum allowed value for a color space. Data will be clipped to be
    # no less than the value of ``min_value``.
    min_value: np.ndarray = np.array([-np.inf])

    # The maximum allowed value for a color space. Data will be clipped to be
    # no more than the value of ``max_value``.
    scale_factor: np.ndarray = np.array([1.])

    __spacename__: str = ''

    def get_data(self) -> np.ndarray:
        """
        The stored color data

        Indexing into the space data instance is equivalent to indexing into 
        the data itself::

            space.data[inds]

        is the same as::

            space[inds]"""
        b_shape = [-1 if k == self.axis else 1
                   for k in range(self.raw_data.ndim)]
        min_value = np.array(self.min_value).reshape(b_shape)
        max_value = np.array(self.max_value).reshape(b_shape)
        d = np.clip(self.raw_data, min_value, max_value)
        return self.scale_factor * d if self.is_scaled else d

    data: np.ndarray = lazy_property(get_data)

    @abstractmethod
    def get_raw_data(self) -> np.ndarray:
        """
        The raw unscaled, un-clipped data. All conversions are done on this
        value.
        
        :return: the raw data 
        """
        pass

    raw_data: np.ndarray = lazy_property(get_raw_data)

    @abstractmethod
    def get_axis(self) -> int:
        """
        The axis in the data array that the color components lie along.
        
        Changing the axis will permute the dimensions of the underlying
        array, so that the color space components lie along the new axis
        """
        pass

    @abstractmethod
    def set_axis(self, a: int):
        pass

    axis: int = lazy_property(get_axis, set_axis)

    @abstractmethod
    def get_illuminant(self) -> Illuminant:
        """
        The illuminant. This combined with the 
        :py:attr:`~ColorSpaceData.observer` 
        determines the reference white point of the space."""
        pass

    @abstractmethod
    def set_illuminant(self, ill: Illuminant):
        pass

    illuminant: Illuminant = lazy_property(get_illuminant, set_illuminant)

    @abstractmethod
    def get_observer(self) -> Observer:
        """
        The observer. This combined with the 
        :py:attr:`~ColorSpaceData.illuminant` 
        determines the reference white point of the space.
        """
        pass

    @abstractmethod
    def set_observer(self, obs: Observer):
        pass

    observer: Observer = lazy_property(get_observer, set_observer)

    @abstractmethod
    def get_rgbs(self) -> RgbSpecification:
        """
        The RGB color space specification.
        """
        pass

    @abstractmethod
    def set_rgbs(self, r: RgbSpecification):
        pass

    rgbs: RgbSpecification = lazy_property(get_rgbs, set_rgbs)

    @abstractmethod
    def get_caa(self) -> ChromaticAdaptationAlgorithm:
        """
        The chromatic adaptation algorithm.
        """
        pass

    @abstractmethod
    def set_caa(self, c: ChromaticAdaptationAlgorithm):
        pass

    caa: ChromaticAdaptationAlgorithm = lazy_property(get_caa, set_caa)

    @abstractmethod
    def get_is_scaled(self) -> bool:
        """
        Whether the data is scaled
        
        If set to ``True``, then the data is scaled by the scale factor of the
        space.
        """
        pass

    @abstractmethod
    def set_is_scaled(self, tf: bool):
        pass

    is_scaled: bool = lazy_property(get_is_scaled, set_is_scaled)

    def get_components(self) -> Tuple[np.ndarray, ...]:
        """
        Tuple containing the correct slices of the data to get the 
        individual color space components. For example, in :class:`LabData`, 
        this property would contain ``(L*, a*, b*)``.

            >>> lab = LabData([[50., 25., 25.], [75., 0., 60.]])
            >>> lab.components[0]
            array([[ 50.],
                   [ 75.]])

        """
        component_inds = construct_component_inds(self.axis,
                                                  self.data.ndim,
                                                  self.num_components,
                                                  min_ndims=0)
        return tuple(self[c] for c in component_inds)

    components: Tuple[np.ndarray, ...] = lazy_property(get_components)

    def get_num_components(self) -> int:
        """
        :return: The number of components in the color space. For example 
           :class:`LabData` has three components: L*, a*, b*. 
        """
        return 3

    num_components: int = lazy_property(get_num_components)

    @abstractmethod
    def to(self, space: Union[str, type]) -> 'ColorSpaceData':
        """
        Convert this space to another space.::
        
            >>> lab = LabData([50., 25., 25.])
            >>> xyz = lab.to(XyzData)
        
        :param space: either the name or the class of the destination color 
           space. 
        :return: The new color space data

        """
        pass

    def get_kwargs(self) -> dict:
        """
        The keyword arguments used to construct the object
        :return: 
        """
        return {'axis': self.axis,
                'illuminant': self.illuminant,
                'observer': self.observer,
                'rgbs': self.rgbs,
                'caa': self.caa,
                'is_scaled': self.is_scaled}

    kwargs: dict = property(get_kwargs)

    def __array__(self, dtype) -> np.ndarray:
        if dtype == self.data.dtype:
            return self.data
        else:
            return np.array(self.data, copy=True, dtype=dtype)

    def __repr__(self) -> str:
        kwarg_repr = ', '.join(f'{key!s}={value!r}'
                               for key, value in self.kwargs.items())
        return f'{type(self).__name__}({self.data!r}, {kwarg_repr})'

    def __getitem__(self, *args, **kwargs) -> Union[float, np.ndarray]:
        return self.data.__getitem__(*args, **kwargs)

    def __eq__(self, other) -> bool:
        return (type(self) == type(other)
                and np.allclose(self.data, other.data)
                and self.axis == other.axis
                and self.illuminant == other.illuminant
                and self.observer == other.observer
                and self.rgbs == other.rgbs
                and self.caa == other.caa
                and self.is_scaled == other.is_scaled)


# A type variable for color space types.
ColorSpace = Type[ColorSpaceData]

# Stores all named color spaces
_space_name_to_type_map: Dict[str, ColorSpace] = {}


def get_space(space: Union[str, ColorSpace]) -> Tuple[str, ColorSpace]:
    """
    Get the space name and class associated with it
    """
    if isinstance(space, str):
        space_name = space
        space_class = get_space_class(space)
    else:
        if not isinstance(space, type):
            space = type(space)
        space_class = space
        space_name = get_space_name(space_class)
    return space_name, space_class


def get_space_class(space_name: str) -> ColorSpace:
    """
    Get the color space class associated with a color space

    :param space_name: The name of the space

    >>> get_space_class('CIEXYZ')
    XyzData
    """
    if isinstance(space_name, str):
        if space_name in _space_name_to_type_map:
            return _space_name_to_type_map[space_name]
        else:
            raise UndefinedColorSpaceError(space_name)
    else:
        raise TypeError('get_space_class expected a str object, but got a '
                        f'{type(space_name).__name__} instead.')


def get_space_name(space_class: ColorSpace) -> str:
    """Get the color space name associated with a color space"""
    if isinstance(space_class, type):
        if hasattr(space_class, '__spacename__') and space_class.__spacename__:
            return space_class.__spacename__
        else:
            raise UndefinedColorSpaceError(space_class)
    else:
        raise TypeError('get_space_name expected a type object, but got a '
                        f'{type(space_class).__name__} instead.')


def color_space(name: str) -> Callable[[ColorSpace], ColorSpace]:
    """
    Decorator that registers a class as a color space.

    :return: decorator that returns the class after registering it

    The ``color_space`` decorator registers a class as a color space for color 
    conversions.::

       @color_space('test1')
       class TestSpaceData(ColorSpaceDataImpl):
           pass

    """

    def decorator(cls: ColorSpace) -> ColorSpace:
        cls.__spacename__ = name
        _space_name_to_type_map[name] = cls
        return cls

    return decorator


# A type variable for things that can be turned into arrays.
ArrayLike = Union[np.ndarray, Iterable['ArrayLike'], float]


class ColorSpaceDataImpl(ColorSpaceData, SetGet):
    """
    A full implementation of the :class:`ColorSpaceDataBase` interface, with 
    internal storage for the data.
    """

    def __init__(self,
                 data: ArrayLike,
                 *,
                 axis: int=None,
                 illuminant: Illuminant=get_default_illuminant(),
                 observer: Observer=get_default_observer(),
                 rgbs: RgbSpecification=get_default_rgb_specification(),
                 caa: ChromaticAdaptationAlgorithm=
                 get_default_chromatic_adaptation_algorithm(),
                 is_scaled: bool=False):
        """
        
        :param data: the color space data to contain
        :param axis: the axis along which the color data lies. If `axis` is not
           specified, then it will be determined automatically by finding the 
           last dimension with the required size.
        :param illuminant: the illuminant
        :param observer: the observer
        :param rgbs: the rgb specification
        :param caa: the chromatic adaptation algorithm
        :param is_scaled: Whether or not the data is scaled
        """
        if is_scaled:
            self._data = np.array(data, copy=True)/self.scale_factor
        else:
            self._data = np.array(data, copy=True)
        self._data.flags.writeable = False
        self._axis = (axis
                      if axis is not None
                      else get_matching_axis(self._data.shape, 3))
        self._illuminant = (illuminant
                            if illuminant is not None
                            else get_default_illuminant())
        self._observer = (observer
                          if observer is not None
                          else get_default_observer())
        self._rgbs = (rgbs
                      if rgbs is not None
                      else get_default_rgb_specification())
        self._caa = (caa
                     if caa is not None
                     else get_default_chromatic_adaptation_algorithm())
        self._is_scaled = is_scaled

    def get_raw_data(self):
        return self._data

    def get_axis(self) -> int:
        return self._axis

    def set_axis(self, a: int):
        if a is None:
            self._axis = get_matching_axis(self.data.shape,
                                           self.num_components)
        elif a != self.axis:
            new_dims = list(range(self.data.ndim))
            new_dims[a] = self.axis
            new_dims[self.axis] = a
            self._data = self._data.transpose(new_dims)
            self._axis = a

    def get_illuminant(self) -> Illuminant:
        return self._illuminant

    def set_illuminant(self, ill: Illuminant):
        self._illuminant = ill

    def get_observer(self) -> Observer:
        return self._observer

    def set_observer(self, obs: Observer):
        self._observer = obs

    def get_rgbs(self) -> RgbSpecification:
        return self._rgbs

    def set_rgbs(self, r: RgbSpecification):
        self._rgbs = r

    def get_caa(self) -> ChromaticAdaptationAlgorithm:
        return self._caa

    def set_caa(self, c: ChromaticAdaptationAlgorithm):
        self._caa = c

    def get_is_scaled(self) -> bool:
        return self._is_scaled

    def set_is_scaled(self, is_scaled: bool) -> None:
        self._is_scaled = is_scaled

    def to(self, space: Union[str, type],
           **kwargs) -> ColorSpaceData:
        from chromathicity.convert import convert
        to_space, to_class = get_space(space)
        from_space = get_space_name(type(self))
        self_kwargs = self.kwargs
        self_is_scaled = self_kwargs.pop('is_scaled')
        converted_data = convert(self.raw_data,
                                 from_space=from_space,
                                 to_space=to_space,
                                 **self_kwargs)
        new_data = to_class(data=converted_data,
                            **self_kwargs)
        new_data.is_scaled = self_is_scaled
        new_data.set(**kwargs)
        return new_data


@color_space(names.REFLECTANCE_SPECTRUM)
class ReflectanceSpectrumData(ColorSpaceDataImpl):
    """
    Contains raw reflectance spectral data
    
    """
    scale_factor = 100

    def __init__(self,
                 data: Union[np.ndarray, Iterable[Any], ColorSpaceData],
                 wavelengths: Union[np.ndarray, Iterable[float]],
                 *,
                 axis=None,
                 **kwargs):
        """
        
        In addition to the usual data arguments, this data also needs the 
        wavelengths of the spectra. 
        
        :param data: The spectral data [reflectance].
        :param wavelengths: The wavelengths that correspond to the spectra.
        :param axis: The axis along which the spectra lie.
        :param illuminant: The illuminant
        :type illuminant: Illuminant
        :param observer: The observer
        :type observer: Observer
        :param rgbs: The RGB specification
        :type rgbs: RgbSpecification
        :param caa: The chromatic adaptation algorithm.
        :type caa: ChromaticAdaptationAlgorithm 
        """
        if not isinstance(data, np.ndarray):
            data = np.array(data)
        if axis is None:
            if isinstance(data, ColorSpaceData):
                if data.axis is None:
                    axis = get_matching_axis(data.data.shape, len(wavelengths))
                else:
                    axis = data.axis
            else:
                axis = get_matching_axis(data.shape, len(wavelengths))
        # noinspection PyArgumentList
        super().__init__(data, axis=axis, **kwargs)
        self._wavelengths = np.array(wavelengths, copy=True)

    def get_num_components(self):
        """The number of wavelengths in the data"""
        return self.wavelengths.size

    def get_wavelengths(self) -> np.ndarray:
        """The wavelengths that correspond to the data"""
        return self._wavelengths

    wavelengths = property(get_wavelengths)

    def get_kwargs(self):
        kwargs = super().get_kwargs()
        kwargs['wavelengths'] = self.wavelengths
        return kwargs


# noinspection PyMethodOverriding
class WhitePointSensitive(ColorSpaceDataImpl):
    """
    This class implements automatic chromatic adaptation whenever the white 
    point of the illuminant/observer combination changes.
    
    Any subclass representing a space that is sensitive to the white point of 
    the illuminant/observer combination should inherit from this class.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def set_illuminant(self, ill: Illuminant):
        self.change_white_point(ill, self._observer)

    def set_observer(self, obs):
        self.change_white_point(self._illuminant, obs)

    def change_white_point(self, illuminant: Illuminant, observer: Observer):
        """
        Change the reference white point of the space by setting a new 
        illuminant and observer
        
        :param illuminant: The new illuminant
        :param observer: The new observer
        :return: self
        """
        from chromathicity.convert import xyz2xyz, convert

        source_white_point = self._illuminant.get_white_point(self._observer)
        destination_white_point = illuminant.get_white_point(observer)
        if np.allclose(source_white_point, destination_white_point):
            self._illuminant = illuminant
            self._observer = observer
            return self

        source_xyz = convert(self._data,
                             from_space=self.__spacename__,
                             to_space='xyz',
                             illuminant=illuminant,
                             observer=observer,
                             rgbs=self._rgbs,
                             caa=self._caa)
        dest_xyz = xyz2xyz(source_xyz,
                           source_white_point=source_white_point,
                           destination_white_point=destination_white_point,
                           axis=self._axis,
                           caa=self._caa)
        self._data = convert(dest_xyz,
                             from_space='xyz',
                             to_space=self.__spacename__,
                             illuminant=illuminant,
                             observer=observer,
                             rgbs=self._rgbs,
                             caa=self._caa)
        self._illuminant = illuminant
        self._observer = observer
        return self


@color_space(names.XYZ)
class XyzData(WhitePointSensitive):
    """
    Represents data from the CIE XYZ color space. 
    
    From `Wikipedia <https://en.wikipedia.org/wiki/CIE_1931_color_space>`_:
    
        *The CIE 1931 color spaces were the first defined quantitative links 
        between physical pure colors (i.e. wavelengths) in the 
        electromagnetic visible spectrum, and physiological perceived colors 
        in human color vision. The mathematical relationships that define 
        these color spaces are essential tools for color management, 
        important when dealing with color inks, illuminated displays, 
        and recording devices such as digital cameras.*
    
        *The CIE 1931 RGB color space and CIE 1931 XYZ color space were 
        created by the International Commission on Illumination (CIE) in 
        1931. They resulted from a series of experiments done in the late 
        1920s by William David Wright and John Guild. The experimental 
        results were combined into the specification of the CIE RGB color 
        space, from which the CIE XYZ color space was derived.* 
    """
    scale_factor = 100


@color_space(names.XYY)
class XyyData(WhitePointSensitive):
    """
    Represents data from the CIE xyY color space
    
    From `Wikipedia <https://en.wikipedia.org/wiki/CIE_1931_color_space#
    CIE_xy_chromaticity_diagram_and_the_CIE_xyY_color_space>`_
    
        Since the human eye has three types of color sensors that respond to 
        different ranges of wavelengths, a full plot of all visible colors is 
        a three-dimensional figure. However, the concept of color can be 
        divided into two parts: brightness and chromaticity. For example, 
        the color white is a bright color, while the color grey is considered 
        to be a less bright version of that same white. In other words, 
        the chromaticity of white and grey are the same while their 
        brightness differs. 

        The CIE XYZ color space was deliberately designed so that the Y 
        parameter is a measure of the luminance of a color. The chromaticity 
        of a color is then specified by the two derived parameters x and y, 
        two of the three normalized values being functions of all three 
        tristimulus values X, Y, and Z.
    """
    pass


@color_space(names.NORMALIZED_XYZ)
class NormalizedXyzData(WhitePointSensitive):
    """
    This space is the CIE XYZ space, normalized by the white point
    """
    pass


@color_space(names.LAB)
class LabData(WhitePointSensitive):
    """
    Represents the CIE L*a*b* color space.
    
    From `Wikipedia <https://en.wikipedia.org/wiki/Lab_color_space#CIELAB>`_:
    
        *CIE L\*a\*b\* (CIELAB) is a color space specified by the International 
        Commission on Illumination (French Commission internationale de 
        l'éclairage, hence its CIE initials). It describes all the colors 
        visible to the human eye and was created to serve as a 
        device-independent model to be used as a reference.* 

        *The three coordinates of CIELAB represent the lightness of the color 
        (L\* = 0 yields black and L\* = 100 indicates diffuse white; specular 
        white may be higher), its position between red/magenta and green (a\*, 
        negative values indicate green while positive values indicate 
        magenta) and its position between yellow and blue (b\*, negative 
        values indicate blue and positive values indicate yellow). The 
        asterisk (\*) after L, a and b are pronounced* star *and are part of the 
        full name, since they represent L\*, a\* and b\*, to distinguish them 
        from Hunter's L, a, and b.*
    """
    min_value = np.array([0., -np.inf, -np.inf])
    max_value = np.array([100., np.inf, np.inf])


@color_space(names.LCH)
class LchData(WhitePointSensitive):
    """
    Represents the CIELCh color space.
     
    From `Wikipedia <https://en.wikipedia.org/wiki/Lab_color_space#
    Cylindrical_representation:_CIELCh_or_CIEHLC>`_:
    
        The CIELCh color space is a CIELab cube color space, where instead of 
        Cartesian coordinates a*, b*, the cylindrical coordinates C* (chroma, 
        relative saturation) and h° (hue angle, angle of the hue in the 
        CIELab color wheel) are specified. The CIELab lightness L* remains 
        unchanged. 
    """
    min_value = np.array([0., 0., 0.])
    max_value = np.array([100., np.inf, 360.])


# noinspection PyMethodOverriding
class RgbsSensitive(WhitePointSensitive):
    """
    This class represents spaces that are sensitive to the choice of the 
    :class:`RgbSpecification`, and will adapt the color data if the 
    :attr:`~RgbsSensitive.rgbs` property value changes.
     
    Any Color space representing RGB data should extend this class.
    """
    scale_factor = 255.
    min_value = np.array([0., 0., 0.])
    max_value = np.array([1., 1., 1.])

    def set_rgbs(self, r: RgbSpecification):
        self.change_rgbs(r, self._caa)

    def set_caa(self, c):
        self.change_rgbs(self._rgbs, c)

    def change_rgbs(self, rgbs, caa):
        from chromathicity.convert import convert

        xyz = convert(self._data,
                      from_space=self.__spacename__,
                      to_space='xyz',
                      axis=self._axis,
                      illuminant=self._illuminant,
                      observer=self._observer,
                      rgbs=self._rgbs,
                      caa=self._caa)
        self._data = convert(xyz,
                             from_space='xyz',
                             to_space=self.__spacename__,
                             axis=self._axis,
                             illuminant=self._illuminant,
                             observer=self._observer,
                             rgbs=rgbs,
                             caa=caa)
        self._rgbs = rgbs
        self._caa = caa


@color_space(names.LINEAR_RGB)
class LinearRgbData(RgbsSensitive):
    """
    Represents data that is uncompanded RGB
    """
    pass


@color_space(names.RGB)
class RgbData(RgbsSensitive):
    pass


@color_space(names.HSL)
class HslData(RgbsSensitive):
    pass


@color_space(names.HSI)
class HsiData(RgbsSensitive):
    pass


@color_space(names.HCY)
class HcyData(RgbsSensitive):
    pass


@color_space(names.HSV)
class HsvData(RgbsSensitive):
    pass



from setuptools import setup

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Programming Language :: Python :: 3.6',
]

KEYWORDS = 'color color-space cielab ciexyz math conversion convert'

setup(
    name='chromathicity',
    version='0.3.0',
    description='Extensible color conversions',
    long_description='',
    url='https://hoogamaphone.github.io/chromathicity/',
    author='Christopher Hoogeboom',
    author_email='chris.hoogeboom@gmail.com',
    license='BSD',
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    packages=['chromathicity'],
    install_requires=['scipy', 'bidict', 'networkx']
)
